import os
import string
from tempfile import NamedTemporaryFile
import numpy as np
from moleculekit.tools.autosegment import autoSegment2
from moleculekit.molecule import Molecule
from moleculekit.writers import _deduce_PDB_atom_name, checkTruncations
from moleculekit.util import ensurelist
import logging


logger = logging.getLogger(__name__)

metal_atypes = ('MG', 'ZN', 'MN', 'CA', 'FE', 'HG', 'CD', 'NI', 'CO', 'CU', 'K', 'LI', 
                'Mg', 'Zn', 'Mn', 'Ca', 'Fe', 'Hg', 'Cd', 'Ni', 'Co', 'Cu', 'Li')


def getPDBQTAtomType(atype, aidx, mol, aromaticNitrogen=False):
    if atype in metal_atypes:  # Save metals from treated as protein atoms
        return atype

    tmptype = ''
    # carbons
    if atype == 'Car':
        tmptype = 'A'
    elif atype.startswith('C'):
        tmptype = 'C'
    # nitrogens
    if atype.startswith('N'):
        tmptype = 'N'
        if atype in ['Nam', 'Npl', 'Ng+']:
            bs, bo = np.where(mol.bonds == aidx)
            if len(bs) == 2:
                tmptype += 'A'
        elif atype == 'Nar':
            # if mol.resname[aidx] == 'HIE':
            bs, bo = np.where(mol.bonds == aidx)
            if len(bs) == 2:
                tmptype += 'a' if aromaticNitrogen else 'A'
            else:
                tmptype += 'n' if aromaticNitrogen else ''
        elif atype[-1] != '+':
            tmptype += 'A'
    # elif atype.startswith('N'):
    #   print(atype, aidx)
    #  tmptype = 'NA'
    # oxygens
    if atype.startswith('O'):
        tmptype = 'OA'
    # sulfurs
    if atype.startswith('S'):
        tmptype = 'S'
        if atype not in ['Sox', 'Sac']:
            tmptype += 'A'

    # hydrogens
    if atype.startswith('H'):
        tmptype = 'H'
        # print(aidx)
        # print(np.where(mol.bonds == aidx))
        try:
            bond = np.where(mol.bonds == aidx)[0][0]
        except Exception as e:
            raise RuntimeError(f'Could not atomtype hydrogen atom with index {aidx} due to no bonding partners.')

        oidx = [a for a in mol.bonds[bond] if a != aidx][0]
        if mol.element[oidx] not in ['C', 'A']:
            tmptype += 'D'

    if tmptype == '':
        tmptype = atype[0]

    return tmptype


def getProperties(mol):
    try:
        import pybel
    except ImportError:
        raise ImportError('Could not import openbabel. The atomtyper requires this dependency so please install it with `conda install openbabel -c acellera`')
    name = NamedTemporaryFile(suffix='.pdb').name
    mol.write(name)
    mpybel = next(pybel.readfile('pdb', name))
    # print(name)
    residues = pybel.ob.OBResidueIter(mpybel.OBMol)
    atoms = [[r.GetName(), r.GetNum(), r.GetAtomID(at), at.GetType(), round(at.GetPartialCharge(), 3)]
             for r in residues
             for at in pybel.ob.OBResidueAtomIter(r)]
    return atoms

def prepareProteinForAtomtyping(mol, guessBonds=True, protonate=True, pH=7, segment=True, verbose=True):
    """ Prepares a Molecule object for atom typing.

    Parameters
    ----------
    mol : Molecule object
        The protein to prepare
    guessBonds : bool
        Drops the bonds in the molecule and guesses them from scratch
    protonate : bool
        Protonates the protein for the given pH and optimizes hydrogen networks
    pH : float
        The pH for protonation
    segment : bool
        Automatically guesses the segments of a protein by using the guessed bonds
    verbose : bool
        Set to False to turn of the printing

    Returns
    -------
    mol : Molecule object
        The prepared Molecule
    """
    from moleculekit.tools.autosegment import autoSegment2

    mol = mol.copy()
    if guessBonds:  # Need to guess bonds at the start for atom selection and for autoSegment
        mol.bondtype = np.array([], dtype=object)
        mol.bonds = mol._guessBonds()

    protsel = mol.atomselect('protein')
    metalsel = mol.atomselect('element {}'.format(' '.join(metal_atypes)))
    notallowed = ~(protsel | metalsel)

    if not np.any(protsel):
        raise RuntimeError('No protein atoms found in Molecule')

    if np.any(notallowed):
        resnames = np.unique(mol.resname[notallowed])
        raise RuntimeError('Found atoms with resnames {} in the Molecule which can cause issues with the voxelization. Please make sure to only pass protein atoms and metals.'.format(resnames))

    protmol = mol.copy()
    protmol.filter(protsel, _logger=False)
    metalmol = mol.copy()
    metalmol.filter(metalsel, _logger=False)

    if protonate:
        from moleculekit.tools.preparation import proteinPrepare
        if np.all(protmol.segid == '') and np.all(protmol.chain == ''):
            protmol = autoSegment2(protmol, fields=('segid', 'chain'), basename='K', _logger=verbose)  # We need segments to prepare the protein
        protmol = proteinPrepare(protmol, pH=pH, verbose=verbose, _loggerLevel='INFO' if verbose else 'ERROR')

    if guessBonds:
        protmol.bonds = protmol._guessBonds()
        # TODO: Should we remove bonds between metals and protein? Should we remove metals before guessing bonds and add them back in? Might crash otherwise?

    if segment:
        protmol = autoSegment2(protmol, fields=('segid', 'chain'), _logger=verbose)  # Reassign segments after preparation

        # Assign separate segment to the metals just in case pybel takes that into account
        if np.any(protmol.chain == 'Z') or np.any(protmol.segid == 'ME'):
            raise AssertionError('Report this issue on the moleculekit github issue tracker. Too many chains in the protein.')
        metalmol.segid[:] = 'ME'
        metalmol.chain[:] = 'Z'
        metalmol.resid[:] = np.arange(metalmol.numAtoms) * 2 + protmol.resid.max() + 1 # Just in case, let's put a residue gap between the metals so that they are considered separate chains no matter what happens

    mol = protmol.copy()
    mol.append(metalmol)
    return mol


def atomtypingValidityChecks(mol):
    logger.info('Checking validity of Molecule before atomtyping. ' \
                'If it gives incorrect results or to improve performance disable it with validitychecks=False. ' \
                'Most of these checks can be passed by using the moleculekit.atomtyper.prepareProteinForAtomtyping function. ' \
                'But make sure you understand what you are doing.')
    protsel = mol.atomselect('protein')
    metals = mol.atomselect('element {}'.format(' '.join(metal_atypes)))
    notallowed = ~(protsel | metals)

    if not np.any(protsel):
        raise RuntimeError('No protein atoms found in Molecule')

    if np.any(notallowed):
        resnames = np.unique(mol.resname[notallowed])
        raise RuntimeError('Found atoms with resnames {} in the Molecule which can cause issues with the voxelization. Please make sure to only pass protein atoms and metals.'.format(resnames))

    if mol.bonds.shape[0] < (mol.numAtoms - 1):
        raise ValueError('The protein has less bonds than (number of atoms - 1). This seems incorrect. You can assign bonds with `mol.bonds = mol._getBonds()`')

    from moleculekit.molecule import calculateUniqueBonds
    uqbonds, _ = calculateUniqueBonds(mol.bonds, mol.bondtype)
    if uqbonds.shape[0] != mol.bonds.shape[0]:
        raise RuntimeError('The protein has duplicate bond information. This will mess up atom typing. Please keep only unique bonds in the molecule. If you want you can use moleculekit.molecule.calculateUniqueBonds for this.')

    if np.all(mol.segid == '') or np.all(mol.chain == ''):
        raise RuntimeError('Please assign segments to the segid and chain fields of the molecule using autoSegment2')

    from moleculekit.tools.autosegment import autoSegment2
    mm = mol.copy()
    mm.segid[:] = ''  # Set segid and chain to '' to avoid name clashes in autoSegment2
    mm.chain[:] = ''
    refmol = autoSegment2(mm, fields=('chain', 'segid'), _logger=False)
    numsegsref = len(np.unique(refmol.segid))
    numsegs = len(np.unique(mol.segid))
    if numsegs != numsegsref:
        raise RuntimeError('The molecule contains {} segments while we predict {}. Make sure you used autoSegment2 on the protein'.format(numsegs, numsegsref))

    if not np.any(mol.element == 'H'):
        raise RuntimeError('No hydrogens found in the Molecule. Make sure to use proteinPrepare before passing it to voxelization. Also you might need to recalculate the bonds after this.')


def getPDBQTAtomTypesAndCharges(mol, aromaticNitrogen=False, validitychecks=True):
    if validitychecks:
        atomtypingValidityChecks(mol)

    atomsProp = getProperties(mol)
    for n, a in enumerate(atomsProp):
        if a[0] == 'HIP':
            if a[2].strip().startswith('C') and a[2].strip() not in ['CA', 'C', 'CB']:
                a[3] = 'Car'

            #print(n, a)
    charges = ['{0:.3f}'.format(a[-1]) for a in atomsProp]
    pdbqtATypes = [getPDBQTAtomType(a[3], n, mol, aromaticNitrogen)
                   for n, a in enumerate(atomsProp)]

    return np.array(pdbqtATypes, dtype='O'), np.array(charges, dtype='float32')


def _getHydrophobic(atypes):
    return atypes == 'C'


def _getAromatic(atypes):
    return (atypes == 'A') | (atypes == 'Na') | (atypes == 'Nn')


def _getAcceptor(atypes):
    return (atypes == 'OA') | (atypes == 'NA') | (atypes == 'SA') | (atypes == 'Na')


def _getDonors(atypes, bonds):
    donors = np.zeros(len(atypes), dtype=bool)
    hydrogens = np.where((atypes == 'HD') | (atypes == 'HS'))[0]
    for h in hydrogens:
        partners = bonds[bonds[:, 0] == h, 1]
        partners = np.hstack((partners, bonds[bonds[:, 1] == h, 0]))
        for p in partners:
            if atypes[p][0] in ('N', 'O', 'S'):
                donors[p] = True
    return donors


def _getPosIonizable(mol):
    # arginine, lysine and histidine
    posIonizables = np.zeros(mol.numAtoms, dtype=bool)

    # ARG
    n_idxs = np.where(((mol.resname == 'ARG') | (mol.resname == 'AR0')) & (
        mol.atomtype == 'N') & (mol.name != 'N'))
    allc_idxs = np.where((mol.resname == 'ARG') & (
        mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)

    aidxs = n_idxs[0].tolist() + c_idxs

    # LYS
    n_idxs = np.where(((mol.resname == 'LYS') | (mol.resname == 'LYN')) & (
        mol.atomtype == 'N') & (mol.name != 'N'))
    aidxs += n_idxs[0].tolist()

    # HIS, HID, HIE, HIP, HSD, HSE
    n_idxs = np.where(((mol.resname == 'HIS') | (mol.resname == 'HID') |
                       (mol.resname == 'HIE') | (mol.resname == 'HIP') |
                       (mol.resname == 'HSE') | (mol.resname == 'HSD') |
                       (mol.resname == 'HSP')) &
                      ((mol.atomtype == 'N') | (mol.atomtype == 'NA') | (mol.atomtype == 'Nn') | (mol.atomtype == 'Na')) &
                      (mol.name != 'N'))

    c_idxs = np.where(((mol.resname == 'HIS') | (mol.resname == 'HID') |
                       (mol.resname == 'HIE') | (mol.resname == 'HIP') |
                       (mol.resname == 'HSE') | (mol.resname == 'HSD') |
                       (mol.resname == 'HSP')) &
                      (mol.atomtype == 'A'))

    aidxs += n_idxs[0].tolist() + c_idxs[0].tolist()

    posIonizables[aidxs] = 1

    return posIonizables


def _getNegIonizable(mol):
    # aspartic and glutamate
    negIonizables = np.zeros(mol.numAtoms, dtype=bool)

    # ASP
    o_idxs = np.where(((mol.resname == 'ASP') | (mol.resname == 'ASH')) &
                      (mol.atomtype == 'OA') & (mol.name != 'O'))
    allc_idxs = np.where(((mol.resname == 'ASP') | (mol.resname == 'ASH')) &
                         (mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)
    aidxs = o_idxs[0].tolist() + c_idxs

    # Glutamate
    o_idxs = np.where(((mol.resname == 'GLU') | (mol.resname == 'GLH')) &
                      (mol.atomtype == 'OA') & (mol.name != 'O'))

    allc_idxs = np.where(((mol.resname == 'GLU') | (mol.resname == 'GLH')) &
                         (mol.atomtype == 'C') & (mol.name != 'C'))[0]
    c_idxs = []
    for c in allc_idxs:
        bs = np.where(mol.bonds == c)[0]
        if len(bs) == 3:
            c_idxs.append(c)
    aidxs += o_idxs[0].tolist() + c_idxs

    negIonizables[aidxs] = 1

    return negIonizables


def _getOccupancy(elements):
    return np.array(elements) != 'H'


def _getMetals(atypes):
    return np.in1d(atypes, metal_atypes)


def getFeatures(mol):
    atypes = mol.atomtype
    elements = [el[0] for el in atypes]

    hydr = _getHydrophobic(atypes)
    arom = _getAromatic(atypes)
    acc = _getAcceptor(atypes)
    don = _getDonors(atypes, mol.bonds)
    pos = _getPosIonizable(mol)
    neg = _getNegIonizable(mol)
    metals = _getMetals(atypes)
    occ = _getOccupancy(elements)

    return np.vstack((hydr, arom, acc, don, pos, neg, metals, occ)).T.copy()


def parallel(func, listobj, n_cpus=-1, *args):
    from tqdm import tqdm
    from joblib import Parallel, delayed
    results = Parallel(n_jobs=n_cpus)(delayed(func)(ob, *args)
                                      for ob in tqdm(listobj))
    return results


import unittest
class _TestAtomTyper(unittest.TestCase):
    def test_preparation(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule, mol_equal
        from os import path
        mol = Molecule(path.join(home(dataDir='test-voxeldescriptors'), '1ATL.pdb'))
        ref = Molecule(path.join(home(dataDir='test-voxeldescriptors'), '1ATL_prepared.pdb'))
        mol2 = prepareProteinForAtomtyping(mol, verbose=False)
   
        assert mol_equal(mol2, ref, exceptFields=('coords',))

if __name__ == '__main__':
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging

logger = logging.getLogger(__name__)


class MetricShell(Projection):
    """ Calculates the density of atoms around other atoms.

    The MetricShell class calculates the density of a set of
    interchangeable atoms in concentric spherical shells around some
    other atoms. Thus it can treat identical molecules (like water or
    ions) and calculate summary values like the changes in water density
    around atoms. It produces a n-by-s dimensional vector where n the
    number of atoms in the first selection and s the number of shells
    around each of the n atoms.

    Parameters
    ----------
    sel1 : str
        Atom selection string for the first set of atoms around which the shells will be calculated.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    sel2 : str
        Atom selection string for the second set of atoms whose density will be calculated in shells around `sel1`.
        See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
    numshells : int, optional
        Number of shells to use around atoms of `sel1`
    shellwidth : int, optional
        The width of each concentric shell in Angstroms
    pbc : bool, optional
        Set to false to disable distance calculations using periodic distances
    gap : int, optional
        Not functional yet
    truncate : float, optional
        Set all distances larger than `truncate` to `truncate`
    """

    def __init__(
        self, sel1, sel2, numshells=4, shellwidth=3, pbc=True, gap=None, truncate=None
    ):
        super().__init__()

        from moleculekit.projections.metricdistance import MetricDistance

        self.symmetrical = sel1 == sel2
        self.metricdistance = MetricDistance(
            sel1=sel1,
            sel2=sel2,
            groupsel1=None,
            groupsel2=None,
            metric="distances",
            threshold=8,
            pbc=pbc,
            truncate=truncate,
        )

        self.numshells = numshells
        self.shellwidth = shellwidth
        self.description = None
        self.shellcenters = None

    def _calculateMolProp(self, mol, props="all"):
        props = (
            ("map", "shellcenters", "shelledges", "shellvol")
            if props == "all"
            else props
        )
        res = {}

        mapping = np.vstack(self.metricdistance.getMapping(mol).atomIndexes)
        if "map" in props:
            res["map"] = mapping
        if "shellcenters" in props:
            res["shellcenters"] = (
                np.unique(mapping[:, 0]) if not self.symmetrical else np.unique(mapping)
            )
        if "shelledges" in props:
            res["shelledges"] = np.arange(
                self.shellwidth * (self.numshells + 1), step=self.shellwidth
            )
        if "shellvol" in props:
            res["shellvol"] = (
                4
                / 3
                * np.pi
                * (res["shelledges"][1:] ** 3 - res["shelledges"][:-1] ** 3)
            )

        return res

    def project(self, mol):
        """ Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.
        kwargs :
            Do not use this argument. Only used for backward compatibility. Will be removed in later versions.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        molprops = self._getMolProp(mol, "all")

        distances = self.metricdistance.project(mol)
        if distances.ndim == 1:
            distances = distances[np.newaxis, :]

        return _shells(
            distances,
            molprops["map"],
            molprops["shellcenters"],
            self.numshells,
            molprops["shelledges"],
            molprops["shellvol"],
            self.symmetrical,
        )

    def getMapping(self, mol):
        """ Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        shellcenters = self._getMolProp(mol, "shellcenters")

        from pandas import DataFrame

        types = []
        indexes = []
        description = []
        for i in shellcenters:
            for n in range(self.numshells):
                types += ["shell"]
                indexes += [i]
                description += [
                    "Density of sel2 atoms in shell {}-{} A centered on atom {} {} {}".format(
                        n * self.shellwidth,
                        (n + 1) * self.shellwidth,
                        mol.resname[i],
                        mol.resid[i],
                        mol.name[i],
                    )
                ]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )


def _shells(
    distances, mapping, shellcenters, numshells, shelledges, shellvol, symmetrical
):
    shellcenters = np.unique(mapping[:, 0]) if not symmetrical else np.unique(mapping)

    shellmetric = np.ones((np.size(distances, 0), len(shellcenters) * numshells)) * -1

    for i in range(len(shellcenters)):
        if symmetrical:
            cols = np.any(mapping == shellcenters[i], axis=1)
        else:
            cols = mapping[:, 0] == shellcenters[i]

        for e in range(len(shelledges) - 1):
            inshell = (distances[:, cols] > shelledges[e]) & (
                distances[:, cols] <= shelledges[e + 1]
            )
            shellmetric[:, (i * numshells) + e] = np.sum(inshell, axis=1) / shellvol[e]

    return shellmetric


import unittest


class _TestMetricShell(unittest.TestCase):
    def test_metricshell(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.pdb")
        )
        mol.read(path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"))

        metr = MetricShell("protein and name CA", "resname MOL and noh")
        data = metr.project(mol)

        refdata = np.load(
            path.join(home(dataDir="test-projections"), "metricshell", "refdata.npy")
        )

        assert np.allclose(data, refdata), "Shell density calculation is broken"

    def test_metricshell_simple(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path
        import numpy as np

        mol = Molecule().empty(3)
        mol.name[:] = "CL"
        mol.resname[:] = "CL"
        mol.resid[:] = np.arange(3)
        mol.coords = np.zeros((3, 3, 1), dtype=np.float32)

        mol.coords[0, :, 0] = [0, 0, 0]
        mol.coords[1, :, 0] = [0.5, 0, 0]
        mol.coords[2, :, 0] = [0, 1.5, 0]

        metr = MetricShell("name CL", "name CL", pbc=False)
        data = metr.project(mol)
        # fmt: off
        refdata = np.array([[0.01768388256576615, 0.0, 0.0, 0.0, 0.01768388256576615, 0.0, 0.0, 0.0, 0.01768388256576615, 0.0, 0.0, 0.0]])
        # fmt: on
        assert np.allclose(data, refdata)

        metr = MetricShell("name CL", "name CL", numshells=2, shellwidth=1, pbc=False)
        data = metr.project(mol)
        refdata = np.array(
            [[0.23873241, 0.03410463, 0.23873241, 0.03410463, 0.0, 0.06820926]]
        )
        assert np.allclose(data, refdata)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from moleculekit.version import version as _version
from moleculekit.home import home as __home
import os
import logging.config

__version__ = _version()

try:
    logging.config.fileConfig(os.path.join(__home(), 'logging.ini'), disable_existing_loggers=False)
except:
    print("MoleculeKit: Logging setup failed")


import urllib.request
import logging
logger = logging.getLogger(__name__)


def _getRCSBtext(url):
    connected = False
    while not connected:
        try:
            response = urllib.request.urlopen(url)
            text = response.read()
        except Exception as coer:
            import time
            logger.warning('Failed to connect to URL {} with error {}. Sleeping 5s and retrying.'.format(url, coer))
            time.sleep(5)
            continue
        connected = True

    return text


def rcsbFindMutatedResidues(pdbid):
    """
    Examples
    --------
    >>> rcsbFindMutatedResidues('3onq')
    {'MSE': 'MET'}
    """
    try:
        from bs4 import BeautifulSoup
        import lxml
    except ImportError:
        raise ImportError('You need to install the \'beautifulsoup4\' and \'lxml\' packages to use this function.')
    tomutate = {}

    url = 'http://www.rcsb.org/pdb/explore.do?structureId={}'.format(pdbid)
    text = _getRCSBtext(url)
    soup = BeautifulSoup(text, 'lxml')
    table = soup.find(id='ModifiedResidueTable')

    if table:
        trs = table.find_all('tr')

        for tr in trs:
            td = tr.find_all('td')
            if td:
                mutname = td[0].find_all('a')[0].text.strip()
                orgname = td[5].text.strip()
                tomutate[mutname] = orgname
    return tomutate


def rcsbFindLigands(pdbid):
    """
    Examples
    --------
    >>> rcsbFindLigands('3onq')
    ['SO4', 'GOL']
    """
    try:
        from bs4 import BeautifulSoup
        import lxml
    except ImportError:
        raise ImportError('You need to install the \'beautifulsoup4\' and \'lxml\' packages to use this function.')
    ligands = []

    url = 'http://www.rcsb.org/pdb/explore.do?structureId={}'.format(pdbid)
    text = _getRCSBtext(url)
    soup = BeautifulSoup(text, 'lxml')
    table = soup.find(id='LigandsTable')
    if table:
        trs = table.find_all('tr')

        for tr in trs:
            td = tr.find_all('td')
            if td:
                name = td[0].find_all('a')[0].text.strip()
                ligands.append(name)
    return ligands


from numba import jit
import numpy as np
from math import sqrt, atan2


@jit('float32(float32, float32)', nopython=True)
def wrapDistance(d, box):
    if box == 0:
        return d
    return d - box * round(d / box)


@jit('float32(float32, float32)', nopython=True)
def wrapBondedDistance(d, box):
    if box == 0:
        return d
    # Assuming bonds can't cross multiple periodic boxes this is marginally faster
    hbox = box / 2
    if d < -hbox:
        return d + box
    elif d > hbox:
        return d - box
    return d

def dihedralAngle(pos, box=None):
    """ Calculates a dihedral angle.

    Parameters
    ----------
    pos: np.ndarray
        An array of 4x3 size where each row are the coordinates of an atom defining the dihedral angle
    box: np.ndarray
        The size of the periodic box

    Returns
    -------
    angle: float
        The angle in radians
    """
    if pos.ndim == 3 and pos.shape[2] > 1:
        if box is None:
            box = np.zeros((3, pos.shape[2]), dtype=pos.dtype)
        return dihedralAngleFrames(pos, box)
    else:
        if box is not None:
            box = box.squeeze()
        return dihedralAngleFull(pos.squeeze(), box)[0]

@jit(nopython=True)
def dihedralAngleFrames(pos, box):
    res = np.zeros(pos.shape[2], dtype=pos.dtype)
    for f in range(pos.shape[2]):
        res[f] = dihedralAngleFull(pos[:, :, f], box[:, f])[0]
    return res

@jit(nopython=True)
def dihedralAngleFull(pos, box=None):
    """ Calculates a dihedral angle.

    Parameters
    ----------
    pos: np.ndarray
        An array of 4x3 size where each row are the coordinates of an atom defining the dihedral angle
    box: np.ndarray
        The size of the periodic box
    """
    if pos.shape[0] != 4 or pos.shape[1] != 3:
        raise RuntimeError('dihedralAngles requires a 4x3 sized coordinate matrix as input.')
    if box is None:
        box = np.zeros(3, dtype=pos.dtype)

    r12 = np.zeros(3)
    r23 = np.zeros(3)
    r34 = np.zeros(3)

    r12[0] = wrapBondedDistance(pos[0, 0] - pos[1, 0], box[0])
    r12[1] = wrapBondedDistance(pos[0, 1] - pos[1, 1], box[1])
    r12[2] = wrapBondedDistance(pos[0, 2] - pos[1, 2], box[2])
    r23[0] = wrapBondedDistance(pos[1, 0] - pos[2, 0], box[0])
    r23[1] = wrapBondedDistance(pos[1, 1] - pos[2, 1], box[1])
    r23[2] = wrapBondedDistance(pos[1, 2] - pos[2, 2], box[2])
    r34[0] = wrapBondedDistance(pos[2, 0] - pos[3, 0], box[0])
    r34[1] = wrapBondedDistance(pos[2, 1] - pos[3, 1], box[1])
    r34[2] = wrapBondedDistance(pos[2, 2] - pos[3, 2], box[2])

    # A = cross(r12, r23)
    A = np.zeros(3)
    A[0] = r12[1] * r23[2] - r12[2] * r23[1]
    A[1] = r12[2] * r23[0] - r12[0] * r23[2]
    A[2] = r12[0] * r23[1] - r12[1] * r23[0]

    # B = cross(r23, r34)
    B = np.zeros(3)
    B[0] = r23[1] * r34[2] - r23[2] * r34[1]
    B[1] = r23[2] * r34[0] - r23[0] * r34[2]
    B[2] = r23[0] * r34[1] - r23[1] * r34[0]

    # C = cross(r23, A)
    C = np.zeros(3)
    C[0] = r23[1] * A[2] - r23[2] * A[1]
    C[1] = r23[2] * A[0] - r23[0] * A[2]
    C[2] = r23[0] * A[1] - r23[1] * A[0]

    rA = 1 / sqrt(A[0] * A[0] + A[1] * A[1] + A[2] * A[2])
    rB = 1 / sqrt(B[0] * B[0] + B[1] * B[1] + B[2] * B[2])
    rC = 1 / sqrt(C[0] * C[0] + C[1] * C[1] + C[2] * C[2])

    B[0] *= rB
    B[1] *= rB
    B[2] *= rB

    cos_phi = (A[0] * B[0] + A[1] * B[1] + A[2] * B[2]) * rA
    sin_phi = (C[0] * B[0] + C[1] * B[1] + C[2] * B[2]) * rC

    phi = -atan2(sin_phi, cos_phi)

    return phi, r12, r23, r34, A, B, C, rA, rB, rC, sin_phi, cos_phi